library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

-- the memory controller unit (mcu) manages the dram. Because the chip will perpeturarly
-- be bandwidth starved, it will basiclly set the timing on everything.
-- memory users (mu) will assert that they need reading/writing done on their behalf, and the mcu
-- will preform the desired operations when it can.  the data for reading/writing will be held
-- in the 'calling' mu.  the calling mu's will share an access buss to the mcu.

entity mcu is
    Port (
   		dram_d_buss		: inout std_logic_vector(31 downto 0);
          dram_a_buss 		: out std_logic_vector(11 downto 0);
          dram_ras_0 		: out std_logic;  -- all the rest of the dram signals are active low
          dram_ras_1 		: out std_logic;
          dram_ras_2 		: out std_logic;
          dram_ras_3 		: out std_logic;
          dram_cas_0 		: out std_logic;
          dram_cas_1 		: out std_logic;
          dram_cas_2 		: out std_logic;
          dram_cas_3 		: out std_logic;
		dram_we			: out std_logic;
		 
		--the mcu's locic interface to the rest of the tvc		 
		tvc_clock		: in		std_logic;
		tvc_reset		: in 	std_logic;

		--memory user access buss 
		mu_buss_address 	: out std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: out  std_logic; --if high the MCU is going to write to a mu
		--if high the MCU expects the selected MU to present its data on the mu_buss_data or to
		--copy data from the mu_buss_data into its internal cache/registers.
		mu_buss_strobe			: out  std_logic;
		
		odd_sc_irq		: in std_logic;
		odd_sc_select		: out std_logic;

		even_sc_irq		: in std_logic;
		even_sc_select		: out std_logic;

		block_reader_irq		: in std_logic;
		block_reader_select	: out std_logic;

		block_writer_irq		: in std_logic;
		block_writer_select	: out std_logic;				 

		rp_writer_irq		: in std_logic;
		rp_writer_select	: out std_logic				 
		);
end mcu;

architecture Behavioral of mcu is
 	TYPE MCU_STATE is (
	MCU_DUMMY_TEST,
	MCU_IDLE,
	MCU_IDLE_REFRESH,
	MCU_XFER_STARTUP,
	MCU_XFER_STARTUP_2,
	MCU_XFER_STARTUP_VALIDATE,
	MCU_SIMPLE_XFER, --four byte aligned start and finish
	MCU_INT_XFER_REFRESH,
	MCU_ENDING_XFER,
	MCU_WAITSTATE );

	signal curr_state : MCU_STATE;
	signal after_waitstate_state : MCU_STATE;

	TYPE REFRESH_STATE is (
	REFRESH_DELAY_AFTER_ROW_ADDRESS,
	REFRESH_DELAY_AFTER_RAS_ASSERT,
	REFRESH_DELAY_AFTER_RAS_DEASSERT );
	signal mcu_refresh_state  : REFRESH_STATE;

	TYPE XFER_STATE is (
 	XFER_DELAY_AFTER_ROW_ADDRESS,
	XFER_DELAY_AFTER_RAS_ASSERT,
	XFER_DELAY_AFTER_COLUMN_ADDRESS,
	XFER_DELAY_AFTER_CAS_ASSERT,
	XFER_DELAY_AFTER_RAS_DEASSERT );
	signal dram_xfer_state  : XFER_STATE;

	--abstractions that let me change the design around quickly
	signal mu_buss_data_in : std_logic_vector(31 downto 0);
	signal mu_buss_data_out : std_logic_vector(31 downto 0);
	signal mu_buss_write_sig : std_logic;
	signal mu_buss_strobe_sig : std_logic;

	signal cas_sig : STD_LOGIC;

	signal c_refresh_row : STD_LOGIC_VECTOR(11 downto 0);
	constant max_refresh_row : STD_LOGIC_VECTOR(11 downto 0) := X"3FF";

	--when this hits ?what?, time to refresh the next row.	 If the mcu is busy with another transfer, interrupt it
	--before the next CAS, do the refresh, then resume the transfer.	 it _could_ be the the same row as we are currently
	--in, but this is not very likly (don't check for eventuality) at a 50mhz clock we should have to refresh every 390 clocks.
	signal refresh_timer : STD_LOGIC_VECTOR(11 downto 0);
	signal refresh_needed : std_logic; 

	signal signal_timer : STD_LOGIC_VECTOR(7 downto 0); -- timing of all dram signals is handled by this timer.

	signal active_mu : STD_LOGIC_VECTOR(2 downto 0); --0 is none 1 is the odd_sc 2 is the even_sc, 3 is the reader 4 is the writer.

	--we service mus in a round-robbin fashion with the exception of the pixel-pusher(s) that have priority
	--we maintain a counter that refers to the next mu to check to see if it needs a data transfer.
	signal next_mu_to_check : STD_LOGIC; -- if 0 check the reader if 1 check the writer.

	--this is the data 'parsed' from the selected mu's request registers.
	--we don't need to store all the values parsed. instead we directly use some of them.
	--the address in ram the mu is having us read/write from. I can only deal with the bottom ?21? bits of it with current hardware
	--signal mu_req_byte_address 	: STD_LOGIC_VECTOR(31 downto 0); (don't need to save this).
 	signal mu_req_count	: STD_LOGIC_VECTOR(23 downto 0); --the number of bytes the mu is having us read/write for it.
	signal mu_req_write	: std_logic;

	signal xfer_prev_dram_address_ras : std_logic_vector(10 downto 0); --we use this to know if we must change RAS lines during a transfer
	signal xfer_curr_dram_address 	: std_logic_vector(31 downto 0); --set to what would have been in mu_req_byte_address
	signal xfer_end_dram_address  	: std_logic_vector(31 downto 0);
	signal xfer_curr_mu_address		: std_logic_vector(11 downto 0); --this is the address of the mu's dword 'buffer' that will be next written or read

	procedure BeginRefreshCycle(	signal mcu_refresh_state : out REFRESH_STATE;
							signal c_refresh_row 	: inout STD_LOGIC_VECTOR(11 downto 0);
							signal signal_timer 	: out STD_LOGIC_VECTOR(7 downto 0);
							signal dram_a_buss 		: out std_logic_vector(11 downto 0) ) is
	begin
		dram_a_buss <= c_refresh_row;
	 	mcu_refresh_state <= REFRESH_DELAY_AFTER_ROW_ADDRESS;
		signal_timer <= X"01";
		if c_refresh_row = max_refresh_row	 then
			c_refresh_row <= X"000";
		else
			c_refresh_row <= c_refresh_row + 1;
		end if;
	end BeginRefreshCycle;

 	--this is a seperate procedure because the functionality is needed in HandleRefresh and HandleXFERStartupValidate
	procedure XFERStart( 	signal curr_state			: inout MCU_STATE;
						signal dram_xfer_state 		: inout XFER_STATE;
						signal signal_timer 		: inout STD_LOGIC_VECTOR(7 downto 0);
						signal xfer_curr_dram_address : in std_logic_vector(31 downto 0);
						signal dram_a_buss 			: out std_logic_vector(11 downto 0) ) is
	begin
		dram_a_buss <= '0' & xfer_curr_dram_address(22 downto 12); --hack. check this value
		signal_timer <= X"01";  --one clock cycle will pass before this is examined again
		curr_state <= MCU_SIMPLE_XFER;
		dram_xfer_state <= XFER_DELAY_AFTER_ROW_ADDRESS;
	end XFERStart;

 	procedure HandleRefresh( signal curr_state			: inout MCU_STATE;
						signal mcu_refresh_state 	: inout REFRESH_STATE;
						signal dram_xfer_state 		: inout XFER_STATE;
						signal signal_timer 		: inout STD_LOGIC_VECTOR(7 downto 0);
						signal xfer_curr_dram_address : in std_logic_vector(31 downto 0);
						signal dram_a_buss 			: out std_logic_vector(11 downto 0);
						signal dram_ras_0 			: out std_logic;
          				signal dram_ras_1 			: out std_logic;
          				signal dram_ras_2 			: out std_logic;
          				signal dram_ras_3 			: out std_logic ) is
	begin
	case mcu_refresh_state is
		when REFRESH_DELAY_AFTER_ROW_ADDRESS =>
			if signal_timer = x"01" then
				dram_ras_0 <= '0';
				dram_ras_1 <= '0';
				dram_ras_2 <= '0';
				dram_ras_3 <= '0';
				signal_timer <= X"01";
				mcu_refresh_state <= REFRESH_DELAY_AFTER_RAS_ASSERT;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when REFRESH_DELAY_AFTER_RAS_ASSERT =>
			if signal_timer = x"04" then
				dram_ras_0 <= '1';
				dram_ras_1 <= '1';
				dram_ras_2 <= '1';
				dram_ras_3 <= '1';
				signal_timer <= X"01";
				mcu_refresh_state <= REFRESH_DELAY_AFTER_RAS_DEASSERT;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when REFRESH_DELAY_AFTER_RAS_DEASSERT =>
			if signal_timer = x"02" then
				if curr_state = MCU_IDLE_REFRESH then --we are done. otherwise we must resume data transfers.
			   		curr_state <= MCU_IDLE; 
				else
					XFERStart( curr_state, dram_xfer_state, signal_timer,  xfer_curr_dram_address, dram_a_buss );
				end if;
			else
				signal_timer <= signal_timer + 1;
			end if;
	 end case;
	end HandleRefresh;

	procedure HandleIdleClock(   	signal curr_state		: inout MCU_STATE;
							signal after_waitstate_state : out MCU_STATE;
							signal mcu_refresh_state : out REFRESH_STATE;
							signal c_refresh_row 	: inout STD_LOGIC_VECTOR(11 downto 0);
							signal refresh_needed	: in std_logic;
							signal signal_timer 	: out STD_LOGIC_VECTOR(7 downto 0);
							signal active_mu		: inout  std_logic_vector(2 downto 0);
							signal dram_a_buss 		: out std_logic_vector(11 downto 0);
							signal mu_buss_address	: out std_logic_vector(11 downto 0);
							signal mu_buss_write_sig	: out  std_logic;
							signal mu_buss_strobe_sig	: out  std_logic;
							signal odd_sc_irq		: in	std_logic;
							signal odd_sc_select	: out std_logic;
							signal even_sc_irq		: in	std_logic;
							signal even_sc_select	: out std_logic;
							signal block_reader_irq	: in	std_logic;
							signal block_reader_select	: out std_logic;
							signal block_writer_irq	: in std_logic;
							signal block_writer_select	: out std_logic;
							signal rp_writer_irq	: in std_logic;
							signal rp_writer_select	: out std_logic ) is
	variable starting_xfer :	boolean;
	begin
		if refresh_needed = '1' then 
			curr_state <= MCU_IDLE_REFRESH;
			BeginRefreshCycle(mcu_refresh_state, c_refresh_row, signal_timer, dram_a_buss);			
		else --see if any mu wants work
			starting_xfer := false;
		 	if odd_sc_irq = '1' then
				active_mu <= "001";
				odd_sc_select <= '1';
				starting_xfer := true;
			elsif even_sc_irq = '1' then
				active_mu <= "010";
				even_sc_select <= '1';
				starting_xfer := true;
			elsif block_reader_irq = '1' then
				active_mu <= "011";
				block_reader_select <= '1';
				starting_xfer := true;
			elsif block_writer_irq = '1' then
				active_mu <= "100";
				block_writer_select <= '1';
				starting_xfer := true;
			elsif rp_writer_irq = '1' then
				active_mu <= "101";
				rp_writer_select <= '1';
				starting_xfer := true;	
			end if; --no change in state

			if starting_xfer = true then
				after_waitstate_state <= MCU_XFER_STARTUP;
				curr_state <= MCU_WAITSTATE;
				mu_buss_address <= X"FFC";
				mu_buss_strobe_sig <= '1';
				mu_buss_write_sig <= '0'; --we want to read an address
			end if;
		end if;
	end HandleIdleClock;

	procedure HandleXFERStartup(	signal curr_state			: inout MCU_STATE;
							signal after_waitstate_state 	: out MCU_STATE;
							signal mu_buss_data_in		: in std_logic_vector(31 downto 0);
							signal mu_req_count			: out STD_LOGIC_VECTOR(23 downto 0);
							signal mu_req_write			: out std_logic;
							signal mu_buss_address		: out std_logic_vector(11 downto 0) ) is --HandleIdle already set up the active mu to transfer the configuration data.
	variable local_dram_address :	std_logic_vector(23 downto 0);
	variable local_dram_count   :	std_logic_vector(11 downto 0);
	begin
		--read current data
		mu_req_write <= mu_buss_data_in(31);
		mu_req_count <= mu_buss_data_in(23 downto 0);
		--if mu_buss_data(23 downto 0) = X"000000" then --there was an error but we will handle it later. an error like this should be very rare/never.
		mu_buss_address <= X"FFD";
		after_waitstate_state <= MCU_XFER_STARTUP_2;
		curr_state <= MCU_WAITSTATE; --it will take two clockcycles before the mu_buss_changes
	end HandleXFERStartup;

	procedure HandleXFERStartup2(	signal curr_state			: inout MCU_STATE;
							signal after_waitstate_state 	: out MCU_STATE;
							--signal xfer_prev_dram_address : std_logic_vector(31 downto 0);
							signal mu_buss_data_in		: in std_logic_vector(31 downto 0);
							signal xfer_curr_dram_address : out std_logic_vector(31 downto 0);
							signal xfer_end_dram_address  : out std_logic_vector(31 downto 0);
							signal xfer_curr_mu_address	: out std_logic_vector(11 downto 0);
							signal mu_buss_strobe_sig		: out std_logic ) is
	begin
		--xfer_prev_dram_adderss <= mu_buss_data; --XFER data handles setting this the first time
		xfer_curr_dram_address <= mu_buss_data_in;
		xfer_end_dram_address <= mu_buss_data_in + mu_req_count; --( X"0000" & mu_req_count );
		xfer_curr_mu_address <= X"000";
		mu_buss_strobe_sig <= '0'; --we need a clock cycle to process this.
		curr_state <= MCU_XFER_STARTUP_VALIDATE;
	end HandleXFERStartup2;

 	procedure HandleXFERStartupValidate(	signal curr_state			: inout MCU_STATE;
									signal after_waitstate_state 	: out MCU_STATE;
						   			signal dram_xfer_state 		: inout XFER_STATE;
									signal signal_timer 		: inout STD_LOGIC_VECTOR(7 downto 0);
									signal xfer_curr_dram_address : in std_logic_vector(31 downto 0);
									signal xfer_end_dram_address  : in std_logic_vector(31 downto 0);
									signal dram_a_buss 			: out std_logic_vector(11 downto 0);
									signal mu_buss_address		: out std_logic_vector(11 downto 0);
									signal mu_buss_data_out		: out std_logic_vector(31 downto 0);
									signal mu_buss_write_sig	: out std_logic;
									signal mu_buss_strobe_sig	: out  std_logic;
									signal odd_sc_select		: out std_logic;
									signal even_sc_select		: out std_logic;
									signal block_reader_select		: out std_logic;
									signal block_writer_select		: out std_logic;
									signal rp_writer_select		: out std_logic ) is
	begin
		if xfer_end_dram_address <= xfer_curr_dram_address or --there was an error. either the number of memory addesses was zero, or there was an overflow
		 xfer_curr_dram_address(1 downto 0) /= X"00" or xfer_end_dram_address(1 downto 0) /= X"00" then	--unaligned reads and writes not implemented currently.
			--I could have a mask which would allow me to ignore functional units.
			--probably a bad idea. functional units should just reset themselves when select goes false.
			--if the user does not see drawn what was commanded, there was an error (or a hardware bug).
		  	odd_sc_select <= '0';
		  	even_sc_select <= '0';
			block_reader_select <= '0';
			block_writer_select <= '0';
			rp_writer_select <= '0';
			mu_buss_strobe_sig <= '0';
			after_waitstate_state <= MCU_IDLE;
			curr_state <= MCU_WAITSTATE; --we need the mu to deassert its irq before we are actually idle.
		else
			--get ready to do much more work
			mu_buss_address <= X"000";
			--mu_buss_data_out <= X"F0F0F0F0";
			--mu_buss_write_sig <= '1';
			--mu_buss_strobe_sig <= '1';
			--after_waitstate_state <= MCU_ENDING_XFER;
			--curr_state <= MCU_WAITSTATE;
			XFERStart(curr_state, dram_xfer_state, signal_timer, xfer_curr_dram_address, dram_a_buss);			
		end if;
	end HandleXFERStartupValidate;

	procedure HandleXFERData(signal curr_state				: inout MCU_STATE;
						signal dram_xfer_state 			: inout XFER_STATE;
						signal mcu_refresh_state 		: out REFRESH_STATE;
						signal c_refresh_row 			: inout STD_LOGIC_VECTOR(11 downto 0);
						signal refresh_needed			: in std_logic;
						signal signal_timer 			: inout STD_LOGIC_VECTOR(7 downto 0);
						signal xfer_prev_dram_address_ras 	: inout std_logic_vector(10 downto 0);
						signal xfer_curr_dram_address 	: inout std_logic_vector(31 downto 0);
						signal xfer_end_dram_address  	: in std_logic_vector(31 downto 0);
						signal xfer_curr_mu_address		: inout std_logic_vector(11 downto 0);
 						signal dram_d_buss				: inout std_logic_vector(31 downto 0);
          				signal dram_a_buss 				: out std_logic_vector(11 downto 0);
          				signal dram_ras_0 				: out std_logic;
          				signal dram_ras_1 				: out std_logic;
          				signal dram_ras_2 				: out std_logic;
          				signal dram_ras_3 				: out std_logic;
          				signal cas_sig 				: out std_logic;
						signal dram_we					: out std_logic;
						signal mu_buss_address 			: out std_logic_vector(11 downto 0);
						signal mu_buss_data_in			: in std_logic_vector(31 downto 0);
						signal mu_buss_data_out			: out std_logic_vector(31 downto 0);
						signal mu_buss_write_sig			: out  std_logic; --if high the MCU is going to write to a mu
						signal mu_buss_strobe_sig			: out  std_logic ) is
	begin
	case dram_xfer_state is
		when XFER_DELAY_AFTER_ROW_ADDRESS =>
			if signal_timer = x"02" then
				if xfer_curr_dram_address(22) = '0' then
					dram_ras_0 <= '0';
					dram_ras_2 <= '0';
					dram_ras_1 <= '1'; --should be no change
					dram_ras_3 <= '1';	--no change
				else
					dram_ras_0 <= '1'; --no change
					dram_ras_2 <= '1'; --no change
					dram_ras_1 <= '0';
					dram_ras_2 <= '0';
				end if;
				if mu_req_write = '1' then --we need this data to be available when we apply the column address
					--we previously set the mu_address in the xfer setup
					--go ahead an ask the mu to present its data.
					mu_buss_address <= xfer_curr_mu_address;
					mu_buss_write_sig <= '0';
					mu_buss_strobe_sig <= '1';
				end if;
				signal_timer <= X"01";
				dram_xfer_state <= XFER_DELAY_AFTER_RAS_ASSERT;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when XFER_DELAY_AFTER_RAS_ASSERT =>
			if signal_timer = x"02" then
				dram_a_buss <= "00" & xfer_curr_dram_address(11 downto 2);
				signal_timer <= X"01";
				dram_xfer_state <= XFER_DELAY_AFTER_COLUMN_ADDRESS;
				--setup the mu to send/recieve its data
				if mu_req_write = '1' then	
					dram_we <= '0';	--I don't know if I should do this here or when I assert the CAS
					dram_d_buss <= mu_buss_data_in;	--get the data presented to the dram so when we assert the cas the dram will see it.
				else
					mu_buss_write_sig <= '1';	 --we want to read from the dram and write to the mu.
					mu_buss_address <= xfer_curr_mu_address; --HACK? duplicate of above?
				end if;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when XFER_DELAY_AFTER_COLUMN_ADDRESS=>
			if mu_req_write = '0' then
				mu_buss_strobe_sig <= '0'; --only necessary when we do a fpm cycle
			end if;
			if signal_timer = x"02" then
				cas_sig <= '0';
				signal_timer <= X"01";
				dram_xfer_state <= XFER_DELAY_AFTER_CAS_ASSERT;
				--compute a the next dram_address. we need this done now so we can use the results in the next state (to know if we should end the xfer
				xfer_prev_dram_address_ras <= xfer_curr_dram_address(22 downto 12);
				xfer_curr_dram_address <= xfer_curr_dram_address + 4;
				xfer_curr_mu_address <= xfer_curr_mu_address + 1;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when XFER_DELAY_AFTER_CAS_ASSERT=>
 			if signal_timer = x"02" then	--if reading then data should be availiable
				if mu_req_write = '1' then
					mu_buss_strobe_sig <= '0'; 
					dram_d_buss <= (others => 'Z'); --stop presenting data to the simm
					--dram_we <= '1'; if we were writing, we will keep writing, don't change this
				else
					mu_buss_data_out <= dram_d_buss;
					--mu_buss_data_out <= X"12345678";
					--case xfer_curr_dram_address(3 downto 2) is
					--when "00" => mu_buss_data <= X"C0C0C0C0";
					--when "01" => mu_buss_data <= X"38383838";
					--when "10" => mu_buss_data <= X"07070707";
					--when "11" => mu_buss_data <=  X"00000000";
					--when others => mu_buss_data <= X"00000000";
					--end case;
					mu_buss_strobe_sig <= '1';
				end if;
				if xfer_curr_dram_address = xfer_end_dram_address or
				   xfer_prev_dram_address_ras /= xfer_curr_dram_address(22 downto 12) or
				   refresh_needed = '1' then --our transfer is 'done' or we must change our RAS
					signal_timer <= X"01";
					dram_xfer_state <= XFER_DELAY_AFTER_RAS_DEASSERT;
					dram_ras_0 <= '1';
					dram_ras_1 <= '1';
					dram_ras_2 <= '1';
					dram_ras_3 <= '1';
					cas_sig <= '1';
					dram_we <= '1'; --we will reset it at our next transfer
				else --we are not done with our transfer _and_ we are staying in the current RAS line. compute and present a new CAS
					dram_a_buss <= "00" & xfer_curr_dram_address(11 downto 2);
					mu_buss_address <= xfer_curr_mu_address;
					signal_timer <= X"01";
					dram_xfer_state <= XFER_DELAY_AFTER_COLUMN_ADDRESS;
					cas_sig <= '1';
				end if;
			else
				signal_timer <= signal_timer + 1;
			end if;
		when XFER_DELAY_AFTER_RAS_DEASSERT =>
			mu_buss_strobe_sig <= '0';
			if signal_timer = x"02" then --we _must_ delay for two cycles.
				if xfer_curr_dram_address = xfer_end_dram_address then --the transfer is done 
					curr_state <= MCU_ENDING_XFER;
				elsif refresh_needed = '1' then
					 curr_state <= MCU_INT_XFER_REFRESH;
					BeginRefreshCycle(mcu_refresh_state, c_refresh_row, signal_timer, dram_a_buss);
				else --changing RAS lines
					dram_a_buss <= xfer_curr_dram_address(23 downto 12);
					signal_timer <= X"00";
					dram_xfer_state <= XFER_DELAY_AFTER_ROW_ADDRESS;
				end if;
			else
				signal_timer <= signal_timer + 1;
			end if;
	 end case;
	 end HandleXFERData;

begin
	mu_buss_write <= mu_buss_write_sig;
	mu_buss_strobe <= mu_buss_strobe_sig;
	mu_buss_data_in <= mu_buss_data;

	process(mu_buss_data_out, mu_buss_write_sig)
	begin
		mu_buss_write <= mu_buss_write_sig;
	  if mu_buss_write_sig = '1' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
	end if;
	end process;

	process(cas_sig) is
	begin
		dram_cas_0 <= cas_sig;
          dram_cas_1 <= cas_sig;
          dram_cas_2 <= cas_sig;
          dram_cas_3 <= cas_sig;
	end process;

	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		dram_d_buss <= (others => 'Z');
		dram_a_buss <= X"000";
		dram_ras_0 <= '1';
		dram_ras_1 <= '1';
		dram_ras_2 <= '1';
		dram_ras_3 <= '1';
      cas_sig <= '1';
 		dram_we <= '1';

		mu_buss_address <= X"000";
		mu_buss_write_sig <= '0';
		mu_buss_strobe_sig <= '0';
		mu_buss_data_out <= X"0F0F0F0F";
		
		odd_sc_select <= '0';
		even_sc_select <= '0';
		block_reader_select <= '0';
		block_writer_select <= '0';
		rp_writer_select <= '0';

		--curr_state <= MCU_DUMMY_TEST;
		curr_state <= MCU_IDLE;
		mcu_refresh_state <= REFRESH_DELAY_AFTER_ROW_ADDRESS;
		dram_xfer_state <= XFER_DELAY_AFTER_ROW_ADDRESS;

		mu_req_count <= X"000000";
	    	mu_req_write <= '0';

 		c_refresh_row <= X"000";
		refresh_needed <= '0';
		refresh_timer <= X"000";
		--refresh_timer <= X"23A";
		c_refresh_row <= X"005";

		active_mu <= "000";
		signal_timer <= X"00";
	else	--not a reset
	  	--set to 760. could be 781! we trigger early because it could be several cycles before we can actually start the refresh.
		--have not counted actual cycles yet
		if refresh_timer = X"2F8" then
			if curr_state /= MCU_INT_XFER_REFRESH and curr_state /= MCU_IDLE_REFRESH then
				refresh_needed <= '1';
			end if;
			refresh_timer <= X"000";
		else
			refresh_timer <= refresh_timer + 1;
		end if;

		case curr_state is
		when MCU_DUMMY_TEST =>
			if signal_timer = X"00" then
				mu_buss_data_out <= X"F0F0F0F0";
				mu_buss_write_sig <= '1';
				signal_timer <= signal_timer + 1;
			elsif signal_timer = X"01" then		
				mu_buss_data_out <= X"0F0F0F0F";
				mu_buss_write_sig <= '1';
				signal_timer <= signal_timer + 1;
			elsif signal_timer = X"02" then
				mu_buss_write_sig <= '0';
				mu_buss_data_out <= X"aFbFcFdF";
				signal_timer <= signal_timer + 1;
			else
				dram_d_buss <= mu_buss_data;
				signal_timer <= X"00";
			end if;
		when MCU_IDLE =>
			HandleIdleClock(curr_state, after_waitstate_state, mcu_refresh_state, c_refresh_row, refresh_needed, signal_timer, active_mu, dram_a_buss,
						 mu_buss_address, mu_buss_write_sig, mu_buss_strobe_sig,
						 odd_sc_irq, odd_sc_select, even_sc_irq, even_sc_select, block_reader_irq, block_reader_select, block_writer_irq, block_writer_select, rp_writer_irq, rp_writer_select);
		when MCU_IDLE_REFRESH =>
			refresh_needed <= '0';
			HandleRefresh( curr_state, mcu_refresh_state, dram_xfer_state, signal_timer, xfer_curr_dram_address, dram_a_buss, dram_ras_0, dram_ras_1, dram_ras_2, dram_ras_3);
		when MCU_INT_XFER_REFRESH =>
			refresh_needed <= '0';
 			HandleRefresh( curr_state, mcu_refresh_state, dram_xfer_state, signal_timer, xfer_curr_dram_address, dram_a_buss, dram_ras_0, dram_ras_1, dram_ras_2, dram_ras_3);
		when MCU_XFER_STARTUP =>
			HandleXFERStartup(curr_state, after_waitstate_state, mu_buss_data_in, mu_req_count, mu_req_write, mu_buss_address);
		when MCU_XFER_STARTUP_2 =>
			HandleXFERStartup2(curr_state, after_waitstate_state, mu_buss_data_in, xfer_curr_dram_address, xfer_end_dram_address, xfer_curr_mu_address, mu_buss_strobe_sig);
		when MCU_XFER_STARTUP_VALIDATE =>
			HandleXFERStartupValidate(curr_state, after_waitstate_state, dram_xfer_state, signal_timer,
								xfer_curr_dram_address, xfer_end_dram_address, dram_a_buss, mu_buss_address, mu_buss_data_out, mu_buss_write_sig, mu_buss_strobe_sig,
								 odd_sc_select, even_sc_select, block_reader_select, block_writer_select, rp_writer_select);
		when MCU_SIMPLE_XFER=>
			HandleXFERData(curr_state, dram_xfer_state, mcu_refresh_state, c_refresh_row, refresh_needed, signal_timer,
						xfer_prev_dram_address_ras, xfer_curr_dram_address, xfer_end_dram_address, xfer_curr_mu_address,
						dram_d_buss, dram_a_buss, dram_ras_0, dram_ras_1, dram_ras_2, dram_ras_3, cas_sig, dram_we,
						mu_buss_address, mu_buss_data_in, mu_buss_data_out, mu_buss_write_sig, mu_buss_strobe_sig);
		when MCU_ENDING_XFER=>
			case active_mu is
				when "001" => odd_sc_select <= '0';
				when "010" => even_sc_select <= '0';
				when "011" => block_reader_select <= '0';
				when "100" => block_writer_select <= '0';
				when "101" => rp_writer_select <= '0';
				when others => null;
			end case;
			active_mu <= "000";
			mu_buss_write_sig <= '0';
			after_waitstate_state <= MCU_IDLE;
			curr_state <= MCU_WAITSTATE;
		when MCU_WAITSTATE=>
			curr_state <= after_waitstate_state;
		when others => null;
		end case;
	end if; --reset
	end if; --clock
	end process;
	
end Behavioral;
