library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity pix_pusher is
    Port (
		mu_buss_address 	: in std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: in std_logic;
		mu_buss_strobe		: in std_logic;

		odd_sc_irq		: out std_logic;
		odd_sc_select		: in std_logic;
		even_sc_irq		: out std_logic;
		even_sc_select		: in std_logic;

    		tvc_clock 		: in 	std_logic;
          tvc_reset 		: in 	std_logic;

          disp_pix_value 	: out 	std_logic_vector(7 downto 0);
          disp_horiz_sync 	: out 	std_logic;
          disp_vert_sync 	: out 	std_logic
		);
end pix_pusher;

architecture Behavioral of pix_pusher is
	type PP_INT_STATE is (PP_PRE_DELAY, PP_STARTUP_DELAY, PP_STARTUP_WAIT, PP_NORMAL);
	signal pp_state : PP_INT_STATE;

    	signal pixel_source_sc			: STD_LOGIC; --we can only use a scanline cache if it is filled, otherwise display our test pattern.
	signal curr_drawn_scanline_num	: std_logic_vector(11 downto 0);  --only valid when v_display_pix is true

	signal hsync_counter 			: STD_LOGIC_VECTOR(15 downto 0);
	signal vsync_counter 			: STD_LOGIC_VECTOR(23 downto 0);	 
	signal v_display_pix			: std_logic; --in a vertical region of the display we should be clocking out pixels
	signal h_display_pix			: std_logic;
	signal pixel_clock				: std_logic;

	signal framebuffer_start_addr			: std_logic_vector(23 downto 0);

	signal even_sc_idle 			: std_logic;
	signal even_sc_start_fill		: std_logic;
	signal even_sc_start_fill_addr	: std_logic_vector(23 downto 0);
	signal even_sc_drain_pixels 		: std_logic; 
	signal even_sc_next_pix_value 	: std_logic_vector(7 downto 0);

	signal odd_sc_idle 				: std_logic;
	signal odd_sc_start_fill			: std_logic;
	signal odd_sc_start_fill_addr		: std_logic_vector(23 downto 0);
	signal odd_sc_drain_pixels 		: std_logic; 
	signal odd_sc_next_pix_value 		: std_logic_vector(7 downto 0);

	--both scanline clocks are hooked up to the same address counter. the scanline
	--caches ignore this if they are not in pixel-drain mode
	signal eo_sc_next_pix_addr : std_logic_vector(11 downto 0);

	component scanline_cache
	Port(
	mu_buss_address 		: in std_logic_vector(11 downto 0);
	mu_buss_data			: inout std_logic_vector(31 downto 0);
	mu_buss_write			: in std_logic;
	mu_buss_strobe			: in std_logic;
	sc_irq				: out std_logic;
	sc_select				: in std_logic;
	tvc_clock 			: in std_logic;
	tvc_reset 			: in std_logic;
	sc_idle				: out std_logic;
	sc_start_fill			: in	std_logic;
	sc_start_fill_addr		: in	std_logic_vector(23 downto 0); 
	sc_drain_pixels		: in std_logic;
	next_pix_addr			: in std_logic_vector(11 downto 0);
	next_pix_val			: out std_logic_vector(7 downto 0)
	);
	end component;

	begin

	odd_sc_sc : scanline_cache
		port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
			tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,	sc_irq=>odd_sc_irq, sc_select=>odd_sc_select,
			 sc_idle=>odd_sc_idle, sc_start_fill=>odd_sc_start_fill, sc_start_fill_addr=>odd_sc_start_fill_addr, 
			 sc_drain_pixels=>odd_sc_drain_pixels, next_pix_addr=>eo_sc_next_pix_addr, next_pix_val=>odd_sc_next_pix_value);

	even_sc_sc : scanline_cache
		port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
			tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,	sc_irq=>even_sc_irq, sc_select=>even_sc_select,
			 sc_idle=>even_sc_idle, sc_start_fill=>even_sc_start_fill, sc_start_fill_addr=>even_sc_start_fill_addr, 
			 sc_drain_pixels=>even_sc_drain_pixels, next_pix_addr=>eo_sc_next_pix_addr, next_pix_val=>even_sc_next_pix_value);


	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		pp_state <= PP_PRE_DELAY;	 --we must wait a clock tick for the scanline caches to recognize we want them to fill
		disp_horiz_sync <= '1';
		disp_vert_sync <= '1';
		disp_pix_value <= X"00";
		hsync_counter <= X"0000";
		v_display_pix <= '0';
		h_display_pix <= '0';
		pixel_clock <= '0';
		framebuffer_start_addr <= X"000000";
		even_sc_start_fill	<= '1';
		even_sc_start_fill_addr	<= X"000000";
		odd_sc_start_fill <= '1';
		odd_sc_start_fill_addr <= X"000280";
		eo_sc_next_pix_addr <= X"000";
		curr_drawn_scanline_num <= X"000";
	else	--not a reset
	case pp_state is
	when PP_PRE_DELAY=>	--null; --trap here
		pp_state <= PP_STARTUP_DELAY;
	when PP_STARTUP_DELAY=>
		pp_state <= PP_STARTUP_WAIT;
		even_sc_start_fill <= '0';
		odd_sc_start_fill <= '0';
	when PP_STARTUP_WAIT=>
		if even_sc_idle = '1' and odd_sc_idle = '1' then
			pp_state <= PP_NORMAL;
		end if;
	when PP_NORMAL =>
		if pixel_clock = '1' then
		  pixel_clock <= '0';
		 else
			pixel_clock <= '1';
			case vsync_counter is
			when X"000000"	=> --start the vsync pulse
				disp_vert_sync <= '0';  
			when X"000640" => --end the vsync pulse 
				disp_vert_sync <= '1'; 
          	when X"002580" => --we are starting drawing the part of the screen where pixels are displayed 
				v_display_pix <= '1';
				curr_drawn_scanline_num <= X"000";	 --it is safe to do this now. the places that change it in hsync case are protected by v_display_pix = '1'
			when X"063ce0" => --stop drawing pixels
				v_display_pix <= '0';
			when others => null;
         		end case;
			
			if vsync_counter =  X"065c20" then  --the end of the full display cycle
				vsync_counter <= X"000000";
			else
	    			vsync_counter <= vsync_counter + 1;
			end if;

		   	if hsync_counter = X"320" then
		   		hsync_counter <= X"0000";
		 	else
			   	hsync_counter <= hsync_counter + 1;
           	end if;
	
		 	case hsync_counter is
			when X"0060" =>
				disp_horiz_sync <= '1';
				--decide not on our pixel source for this scanline.  HACK. we should set an invalid flag in the scanline cache for when it does complete
				--we do not display its data on the incorrect display line.  OR.. with properly designed hardware it would never happen, so we should
				--ignore the possibility (just like this test below)
				if v_display_pix = '1' then
					eo_sc_next_pix_addr <= X"000";  --ok to set here. this is protected from change when scanning a scanline by h_display_pix.
					if curr_drawn_scanline_num(0) = '1' and odd_sc_idle = '1' then
						pixel_source_sc <= '1';
						--tell the odd sc to prepare and present its first pixel value.
						odd_sc_drain_pixels	<= '1';	 
					elsif curr_drawn_scanline_num(0) = '0' and even_sc_idle = '1' then
						pixel_source_sc <= '1';
						--tell the even sc to prepare and present its first pixel value
						even_sc_drain_pixels <= '1'; 
					else
						pixel_source_sc <= '0';
					end if;
				end if;	 
			when X"0070" =>
				h_display_pix <= '1';
			when X"02F0" =>
				h_display_pix <= '0';
				if v_display_pix = '1' then 
					if curr_drawn_scanline_num(0) = '1' then --an odd scanline
						if curr_drawn_scanline_num = 479 then
							odd_sc_start_fill_addr <=  X"000280";
						else
							odd_sc_start_fill_addr <= odd_sc_start_fill_addr + 1280; 
						end if;
						odd_sc_drain_pixels <= '0';
						odd_sc_start_fill <= '1';   --it will take the scanline cache several cycles to change modes from drain to full
					else	--an even scanline
						if curr_drawn_scanline_num = 478 then
							even_sc_start_fill_addr <= X"000000";
						else
							even_sc_start_fill_addr <= even_sc_start_fill_addr + 1280;
						end if;
						even_sc_drain_pixels <= '0'; 
						even_sc_start_fill <= '1';   --it will take the scanline cache several cycles to change modes from drain to full
					end if;
				end if; --v_display_pix = '1'
			when X"02F4" =>
				if v_display_pix = '1' then
					if curr_drawn_scanline_num(0) = '1' then
						odd_sc_start_fill <= '0';
					else
						even_sc_start_fill <= '0';
					end if;
					curr_drawn_scanline_num <= curr_drawn_scanline_num + 1;
				end if;
			when X"0320" =>
				disp_horiz_sync <= '0';
			when others => null;
			end case;

-- ************ output pixel values ****************
			if v_display_pix = '1' and h_display_pix = '1' then
				if pixel_source_sc = '1' then
					if curr_drawn_scanline_num(0) = '1' then
				   		disp_pix_value <= odd_sc_next_pix_value;
					else
						disp_pix_value <= even_sc_next_pix_value;
					end if;
				else	  --old pattern generator code
					case eo_sc_next_pix_addr(5 downto 4) is
					when "00" => 	disp_pix_value <= X"c0";
					when "01" =>	disp_pix_value <= X"38";
					when "10" =>  	disp_pix_value <= X"07";
					when others =>	disp_pix_value <= X"00";
					end case;
				end if;
				eo_sc_next_pix_addr <= eo_sc_next_pix_addr + 1;
			else
			  	disp_pix_value <= X"00";	 --?does a high pass filter exist on the monitor inputs?
			end if;
   		end if; --pixel clock
	end case;
	end if;
	end if;
	end process;

end Behavioral;
