library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity scanline_cache is
	Port(
	--memory user access buss 
	mu_buss_address 		: in std_logic_vector(11 downto 0);
	mu_buss_data			: inout std_logic_vector(31 downto 0);
	mu_buss_write			: in  std_logic;
	mu_buss_strobe			: in  std_logic;

	sc_irq				: out std_logic;
	sc_select				: in std_logic;

	tvc_clock 			: in 	std_logic;
	tvc_reset 			: in 	std_logic;

	sc_idle				: out	std_logic; --only true when err... idle.
	sc_start_fill			: in		std_logic; --change the state to ask the mcu with data starting from the address in sc_start_fill_addr
	sc_start_fill_addr		: in		std_logic_vector(23 downto 0); 

	sc_drain_pixels		: in std_logic; --when we are idle and first see this go high, we enter the SC_PIXEL_DRAIN state. we exit this state when it goes to '0'
	next_pix_addr			: in std_logic_vector(11 downto 0); --values must come in sequentially.
	next_pix_val			: out std_logic_vector(7 downto 0)-- when in drain mode this is 'updated' on every clock.
	);
end scanline_cache;

architecture Behavioral of scanline_cache is

	type SC_CURR_STATE is (SC_IDLE_STATE, SC_MCU_DRIVEN, SC_PIXEL_DRAIN);
	signal curr_state : SC_CURR_STATE;
	 
	component RAMB4_S16
	port (WE, EN, RST, CLK: in STD_LOGIC;
	ADDR: in STD_LOGIC_VECTOR(7 downto 0);
	DI: in STD_LOGIC_VECTOR(15 downto 0);
	DO: out STD_LOGIC_VECTOR(15 downto 0));
	end component;
	--signals for the block ram
	signal brWE 	: std_logic;
	signal brEN  	: std_logic;
	signal brRST	: std_logic;

	signal brDI	: std_logic_vector(31 downto 0);
	signal brDO	: std_logic_vector(31 downto 0);

	signal brADDR		: std_logic_vector(7 downto 0);
	--end signals for the block ram

	signal scanline_addr_start 	: std_logic_vector(23 downto 0); --a local copy of the data from sc_start_fill_addr
	signal prev_br_addr_last_pix_value	: std_logic_vector(7 downto 0); --a register that stores the data from the last br address access
	signal prev_br_addr_last_pix_value_valid : std_logic; --when this is true use this value when writing out pixels

 	signal mu_buss_data_in : std_logic_vector(31 downto 0);
	signal mu_buss_data_out : std_logic_vector(31 downto 0);

	signal mcu_selected : std_logic;

	procedure HandleMCUDrivenClock(signal curr_state 		: inout SC_CURR_STATE;
							signal brWE 			: out std_logic;
							signal brEN			: out std_logic;
							signal brDI			: out std_logic_vector(31 downto 0);
							signal brADDR			: out std_logic_vector(7 downto 0);
							signal mcu_selected 	: inout std_logic;
							signal mu_buss_address 	: in std_logic_vector(11 downto 0);
							signal mu_buss_data_in	: in std_logic_vector(31 downto 0);
							signal mu_buss_data_out	: out std_logic_vector(31 downto 0);
							signal mu_buss_write	: in  std_logic;
							signal mu_buss_strobe	: in  std_logic;
							signal sc_irq			: out std_logic) is
	begin
	if sc_select = '0' and mcu_selected = '1' then
		curr_state <= SC_IDLE_STATE;
		sc_irq <= '0';
		brWE <= '0'; --necessary?
		--brEN <= '0';
	elsif sc_select = '1' then
		mcu_selected <= '1';
		if mu_buss_strobe = '1' then
			if mu_buss_write = '1' then
				brADDR <= mu_buss_address(7 downto 0);
				--brDI <= X"c0380700"; --mu_buss_data_in;
				brDI <= mu_buss_data_in;
				brWE <= '1';
			else
				brWE <= '0';
				if mu_buss_address = X"FFC" then
  					mu_buss_data_out <=  X"00" & X"000280";
				elsif mu_buss_address = X"FFD" then
					mu_buss_data_out <= X"00" & scanline_addr_start;
				else
					mu_buss_data_out <= X"0000_0000"; --a problem						
				end if;
			end if; --buss_write
		else --buss_strobe
			brWE <= '0';
		end if;
	end if;
	end HandleMCUDrivenClock;

	procedure HandlePixelDrainClock(signal curr_state 	: inout SC_CURR_STATE;
							signal brEN			: out std_logic;
							signal brDO			: in std_logic_vector(31 downto 0);
							signal brADDR			: out std_logic_vector(7 downto 0);
							signal prev_br_addr_last_pix_value			: inout std_logic_vector(7 downto 0); 
							signal prev_br_addr_last_pix_value_valid 	: inout std_logic;
							signal sc_drain_pixels	: in std_logic;
							signal next_pix_addr	: in std_logic_vector(11 downto 0);
							signal next_pix_val	: out std_logic_vector(7 downto 0) ) is
	begin
	if sc_drain_pixels = '0' then
		curr_state <= SC_IDLE_STATE;
		--brEN <= '0';
	else		
		case next_pix_addr(1 downto 0) is
		when "00" =>
			prev_br_addr_last_pix_value_valid <= '0';
			next_pix_val <= brDO(7 downto 0);
			--next_pix_val <= X"C0";
		when "01" =>
			next_pix_val <= brDO(15 downto 8);
			--next_pix_val <= X"C0";
		when "10" =>
			next_pix_val <= brDO(23 downto 16);
			--next_pix_val <= X"C0";
		when "11" =>
			if prev_br_addr_last_pix_value_valid = '0' then
				prev_br_addr_last_pix_value_valid <= '1';
				prev_br_addr_last_pix_value <= brDO(31 downto 24);
				next_pix_val <= brDO(31 downto 24);
				brADDR <= next_pix_addr(9 downto 2) + 1; --by the very next tvc clock, the data will be on the output
			else
				next_pix_val <= prev_br_addr_last_pix_value;
			end if;
		when others =>
			null; --should never happen
		end case;
	end if;
	end HandlePixelDrainClock;
		
	begin
	high_word : RAMB4_S16
		port map(WE=>brWE, EN=>brEN, RST=>brRST, CLK=>tvc_clock,
				ADDR=>brADDR, DI=>brDI(31 downto 16), DO=>brDO(31 downto 16) );
	low_word	: RAMB4_S16
 		port map(WE=>brWE, EN=>brEN, RST=>brRST, CLK=>tvc_clock,
				ADDR=>brADDR, DI=>brDI(15 downto 0), DO=>brDO(15 downto 0) );				

	process(curr_state)
	begin
		if curr_state = SC_IDLE_STATE then
			sc_idle <= '1';
		else
			sc_idle <= '0';
		end if;
	end process;

	mu_buss_data_in <= mu_buss_data;
	process(mu_buss_data_out, sc_select, mu_buss_strobe, mu_buss_write)
	begin
		if sc_select = '1' and mu_buss_strobe = '1' and mu_buss_write = '0' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
		end if;
	end process;

 	process(tvc_clock)
	variable next_pix_counter : std_logic_vector(11 downto 0);
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		curr_state <= SC_IDLE_STATE;
		brRST <= '0';
		brEN <= '1';
		brWE <= '0';
		brDI <= X"0000_0000";
		sc_irq <= '0';
		next_pix_val <= X"00";
	else
		brRST <= '0';
		case curr_state is
		when SC_IDLE_STATE =>
			if sc_start_fill = '1' then
				scanline_addr_start <= sc_start_fill_addr;
				curr_state <= SC_MCU_DRIVEN;
				mcu_selected <= '0';
				sc_irq <= '1';
				brEN <= '1';
			elsif sc_drain_pixels = '1' then
				curr_state <= SC_PIXEL_DRAIN;
				brEN <= '1';
				brWE <= '0'; --I hope unnecessary
				brADDR <= X"00";
			end if;
		when SC_MCU_DRIVEN =>
			HandleMCUDrivenClock(curr_state, brWE, brEN, brDI, brADDR, mcu_selected,
							mu_buss_address, mu_buss_data_in, mu_buss_data_out, mu_buss_write, mu_buss_strobe, sc_irq);
		when SC_PIXEL_DRAIN =>
			HandlePixelDrainClock(curr_state, brEN, brDO, brADDR, prev_br_addr_last_pix_value, prev_br_addr_last_pix_value_valid,
							sc_drain_pixels, next_pix_addr, next_pix_val );
		end case;
	end if;
	end if;
	end process;
end Behavioral;
